import os
import asyncio
import traceback
import sys
import enum
import contextlib
from typing import List, Any

import asyncio_pool
import unsync

import buelon.core.step
import buelon.hub
import buelon.bucket
import buelon.helpers.json_parser

try:
    import dotenv
    dotenv.load_dotenv()
except ModuleNotFoundError:
    pass

WORKER_HOST = os.environ.get('PIPE_WORKER_HOST', 'localhost')
WORKER_PORT = int(os.environ.get('PIPE_WORKER_PORT', 65432))
PIPE_WORKER_SUBPROCESS_JOBS = os.environ.get('PIPE_WORKER_SUBPROCESS_JOBS', 'false')

bucket_client = buelon.bucket.Client()
hub_client: buelon.hub.HubClient = buelon.hub.HubClient(WORKER_HOST, WORKER_PORT)

JOB_CMD = f'{sys.executable} -c "import buelon.worker;buelon.worker.job()"'


class HandleStatus(enum.Enum):
    success = 'success'
    pending = 'pending'
    almost = 'almost'
    none = 'none'


@contextlib.contextmanager
def new_client_if_subprocess():
    global hub_client
    if PIPE_WORKER_SUBPROCESS_JOBS == 'true':
        yield bucket_client
    else:
        yield hub_client


def job(step_id: str | None = None) -> None:
    if step_id:
        _step = buelon.hub.get_step(step_id)
    else:
        _step = buelon.hub.get_step(os.environ['STEP_ID'])
    print('handling', _step.name)
    try:
        args = [buelon.hub.get_data(_id) for _id in _step.parents]
        r: buelon.core.step.Result = _step.run(*args)
        buelon.hub.set_data(_step.id, r.data)

        with new_client_if_subprocess() as client:
            if r.status == buelon.core.step.StepStatus.success:
                client.done(_step.id)
            elif r.status == buelon.core.step.StepStatus.pending:
                client.pending(_step.id)
            elif r.status == buelon.core.step.StepStatus.reset:
                client.reset(_step.id)
            elif r.status == buelon.core.step.StepStatus.cancel:
                client.cancel(_step.id)
            else:
                raise Exception('Invalid step status')
    except Exception as e:
        print(' - Error - ')
        print(str(e))
        traceback.print_exc()
        with new_client_if_subprocess() as client:
            client.error(
                _step.id,
                str(e),
                f'{traceback.format_exc()}'
            )


async def run(step_id: str | None = None) -> None:
    if PIPE_WORKER_SUBPROCESS_JOBS != 'true':
        return job(step_id)
    env = {**os.environ, 'STEP_ID': step_id}
    p = await asyncio.create_subprocess_shell(JOB_CMD, env=env)
    await p.wait()


async def work():
    _scopes: str = os.environ.get('PIPE_WORKER_SCOPES', 'default')
    scopes: List = _scopes.split(',')
    print('scopes', scopes)

    last_loop_had_steps = True

    async with asyncio_pool.AioPool(size=50) as pool:
        while True:
            steps = hub_client.get_steps(scopes)

            if not steps:
                if last_loop_had_steps:
                    last_loop_had_steps = False
                    print('waiting..')
                await asyncio.sleep(1.)
                continue

            last_loop_had_steps = True

            for s in steps:
                await pool.spawn(run(s))


def main():
    asyncio.run(_main())


async def _main():
    global hub_client
    with hub_client:
        await work()


try:
    from cython.c_worker import *
except (ImportError, ModuleNotFoundError):
    pass


if __name__ == '__main__':
    main()


#
#
# """Worker module for handling pipeline steps.
#
# This module contains functions and classes for managing and executing pipeline steps,
# communicating with a pipeline server, and handling various step statuses.
#
# Typical usage example:
#
#   `python3 worker.py`
#
#   # or
#
#   import worker
#   worker.main()
#
# """
#
# import os
# import asyncio
# import traceback
# import socket
# import sys
# import enum
# from typing import List, Any
#
# import asyncio_pool
#
# import buelon.core.step
# import buelon.hub
# import buelon.bucket
# import buelon.helpers.json_parser
#
# try:
#     import dotenv
#
#     dotenv.load_dotenv()
# except ModuleNotFoundError:
#     pass
#
# WORKER_HOST = os.environ.get('PIPE_WORKER_HOST', 'localhost')
# WORKER_PORT = int(os.environ.get('PIPE_WORKER_PORT', 65432))
# PIPE_WORKER_SUBPROCESS_JOBS = os.environ.get('PIPE_WORKER_SUBPROCESS_JOBS') == 'true'
#
# bucket_client = buelon.bucket.Client()
#
# JOB_CMD = f'{sys.executable} -c "import worker;worker.job()"'
#
#
# class HandleStatus(enum.Enum):
#     """Enumeration of possible handling statuses."""
#     success = 'success'
#     pending = 'pending'
#     almost = 'almost'
#     none = 'none'
#
#
# def job(step_id: str | None = None) -> None:
#     """Execute a job for a given step.
#
#     Args:
#         step_id: The ID of the step to execute. If None, it's read from the environment.
#
#     Raises:
#         Exception: If an invalid step status is encountered.
#     """
#     if step_id:
#         _step = buelon.hub.get_step(step_id)
#     else:
#         _step = buelon.hub.get_step(os.environ['STEP_ID'])
#     print('handling', _step.name)
#     try:
#         args = [buelon.hub.get_data(_id) for _id in _step.parents]
#         r: buelon.core.step.Result = _step.run(*args)
#         buelon.hub.set_data(_step.id, r.data)
#
#         if r.status == buelon.core.step.StepStatus.success:
#             request_done(_step.id)
#         elif r.status == buelon.core.step.StepStatus.pending:
#             request_pending(_step.id)
#         elif r.status == buelon.core.step.StepStatus.reset:
#             request_reset(_step.id)
#         elif r.status == buelon.core.step.StepStatus.cancel:
#             request_cancel(_step.id)
#         else:
#             raise Exception('Invalid step status')
#     except Exception as e:
#         print(' - Error - ')
#         print(str(e))
#         traceback.print_exc()
#         request_error(
#             _step.id,
#             str(e),
#             f'{traceback.format_exc()}'
#         )
#
#
# async def run(step_id: str | None = None) -> None:
#     """Run a job asynchronously.
#
#     Args:
#         step_id: The ID of the step to run. If None, it's read from the environment.
#     """
#     if PIPE_WORKER_SUBPROCESS_JOBS != 'true':
#         return job(step_id)
#     env = {**os.environ, 'STEP_ID': step_id}
#     p = await asyncio.create_subprocess_shell(JOB_CMD, env=env)
#     await p.wait()
#
#
# async def work():
#     """Main worker loop that continuously processes steps."""
#     _scopes: str = os.environ.get('PIPE_WORKER_SCOPES', 'default')
#     scopes: List = _scopes.split(',')
#     print('scopes', scopes)
#
#     last_loop_had_steps = True  # used to print 'waiting..' only once when no steps
#
#     async with asyncio_pool.AioPool(size=50) as pool:
#         while True:
#             steps = request_steps(scopes)
#
#             if not steps:
#                 if last_loop_had_steps:
#                     last_loop_had_steps = False
#                     print('waiting..')
#                 await asyncio.sleep(1.)
#                 continue
#
#             last_loop_had_steps = True
#
#             for s in steps:
#                 await pool.spawn(run(s))
#
#
# def receive(conn) -> bytes:
#     """Receive data from a socket connection.
#
#     Args:
#         conn: The socket connection.
#
#     Returns:
#         The received data as bytes.
#     """
#     data = b''
#     while not data.endswith(b'[-_-]'):
#         v = conn.recv(1024)
#         data += v
#     return data[:-5]
#
#
# def send(conn, data: bytes) -> None:
#     """Send data through a socket connection.
#
#     Args:
#         conn: The socket connection.
#         data: The data to send.
#     """
#     conn.sendall(data + b'[-_-]')
#
#
# def request_steps(scopes: list[str]) -> Any:
#     """Request steps from the buelon.hub server.
#
#     Args:
#         scopes: A list of scopes to request steps for.
#
#     Returns:
#         The steps received from the server.
#     """
#     global WORKER_HOST, WORKER_PORT
#     with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
#         s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
#         s.connect((WORKER_HOST, WORKER_PORT))
#         data = b'get-steps' + buelon.hub.PIPELINE_SPLIT_TOKEN + buelon.helpers.json_parser.dumps(scopes)
#         send(s, data)
#         data = receive(s)
#         return buelon.helpers.json_parser.loads(data)
#
#
# def request_done(step_id: str) -> None:
#     """Send a 'done' request to the buelon.hub server.
#
#     Args:
#         step_id: The ID of the completed step.
#     """
#     global WORKER_HOST, WORKER_PORT
#     with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
#         s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
#         s.connect((WORKER_HOST, WORKER_PORT))
#         data = b'done' + buelon.hub.PIPELINE_SPLIT_TOKEN + step_id.encode()
#         send(s, data)
#
#
# def request_pending(step_id: str) -> None:
#     """Send a 'pending' request to the buelon.hub server.
#
#     Args:
#         step_id: The ID of the pending step.
#     """
#     global WORKER_HOST, WORKER_PORT
#     with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
#         s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
#         s.connect((WORKER_HOST, WORKER_PORT))
#         data = b'pending' + buelon.hub.PIPELINE_SPLIT_TOKEN + step_id.encode()
#         send(s, data)
#
#
# def request_cancel(step_id: str) -> None:
#     """Send a 'cancel' request to the buelon.hub server.
#
#     Args:
#         step_id: The ID of the step to cancel.
#     """
#     global WORKER_HOST, WORKER_PORT
#     with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
#         s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
#         s.connect((WORKER_HOST, WORKER_PORT))
#         data = b'cancel' + buelon.hub.PIPELINE_SPLIT_TOKEN + step_id.encode()
#         send(s, data)
#
#
# def request_reset(step_id: str) -> None:
#     """Send a 'reset' request to the buelon.hub server.
#
#     Args:
#         step_id: The ID of the step to reset.
#     """
#     global WORKER_HOST, WORKER_PORT
#     with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
#         s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
#         s.connect((WORKER_HOST, WORKER_PORT))
#         data = b'reset' + buelon.hub.PIPELINE_SPLIT_TOKEN + step_id.encode()
#         send(s, data)
#
#
# def request_error(step_id: str, msg: str, trace: str) -> None:
#     """Send an 'error' request to the buelon.hub server.
#
#     Args:
#         step_id: The ID of the step that encountered an error.
#         msg: The error message.
#         trace: The error traceback.
#     """
#     global WORKER_HOST, WORKER_PORT
#     with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
#         s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
#         s.connect((WORKER_HOST, WORKER_PORT))
#         data = b'error' + buelon.hub.PIPELINE_SPLIT_TOKEN + buelon.helpers.json_parser.dumps(
#             {'step_id': step_id, 'msg': msg, 'trace': trace})
#         send(s, data)
#
#
# def main():
#     """Main function to start the worker."""
#     asyncio.run(work())
#
#
# try:
#     from cython.c_worker import *
# except (ImportError, ModuleNotFoundError):
#     pass
#
#
# if __name__ == '__main__':
#     main()
#



"""
This module implements a socket server for sending and receiving byte data using keys.

The server allows clients to set, get, and delete data associated with specific keys.
Data is stored in files within a '.bucket' directory.
"""

import socket
import threading
import os
import sys

import psutil

try:
    import dotenv
    dotenv.load_dotenv()
except ModuleNotFoundError:
    pass


# Environment variables for client and server configuration
BUCKET_CLIENT_HOST: str = os.environ.get('BUCKET_CLIENT_HOST', 'localhost')
BUCKET_CLIENT_PORT: int = int(os.environ.get('BUCKET_CLIENT_PORT', 61535))

BUCKET_SERVER_HOST: str = os.environ.get('BUCKET_SERVER_HOST', '0.0.0.0')
BUCKET_SERVER_PORT: int = int(os.environ.get('BUCKET_SERVER_PORT', 61535))

BUCKET_END_TOKEN = b'[-_-]'
BUCKET_SPLIT_TOKEN = b'[*BUCKET_SPLIT_TOKEN*]'

save_path = '.bucket'

database: dict[str, bytes] = {}
database_keys_in_order = []
MAX_DATABASE_SIZE: int = min(1024 * 1024 * 1024 * 1, int(psutil.virtual_memory().total / 8))


if not os.path.exists(save_path):
    os.makedirs(save_path)


def receive(conn: socket.socket, size: int = 1024) -> bytes:
    """
    Receive data from a socket connection until the end marker is found.

    Args:
        conn (socket.socket): The socket connection to receive data from.
        size (int, optional): The maximum number of bytes to receive at once. Defaults to 1024.

    Returns:
        bytes: The received data without the end marker.
    """
    data = b''
    while not data.endswith(BUCKET_END_TOKEN):
        v = conn.recv(size)
        data += v
    token_len = len(BUCKET_END_TOKEN)
    return data[:-token_len]


def send(conn: socket.socket, data: bytes) -> None:
    """
    Send data through a socket connection with an end marker.

    Args:
        conn (socket.socket): The socket connection to send data through.
        data (bytes): The data to be sent.
    """
    conn.sendall(data + BUCKET_END_TOKEN)


def retry_connection(func: callable):

    """
    Decorator to retry a function call if a ConnectionResetError occurs.

    Args:
        func (callable): The function to be decorated.

    Returns:
        callable: The decorated function.
    """
    def wrapper(*args, **kwargs):
        tries = 4
        kwargs['timeout'] = kwargs.get('timeout', 60 * 5.)
        for i in range(tries):
            try:
                return func(*args, **kwargs)
            except (TimeoutError, ConnectionResetError):
                kwargs['timeout'] *= 2
        raise
    return wrapper


class Client:
    """A client for interacting with the bucket server."""
    PORT: int
    HOST: str

    def __init__(self):
        """Initialize the client with host and port from environment variables."""
        self.PORT = BUCKET_CLIENT_PORT
        self.HOST = BUCKET_CLIENT_HOST

    @retry_connection
    def set(self, key: str, data: bytes, timeout: float | None = 60 * 5.) -> None:
        """
        Set data for a given key on the server.

        Args:
            key (str): The key to associate with the data.
            data (bytes): The data to store.
            timeout (float, optional): The timeout for the socket connection in seconds. Defaults to 60 * 5.
        """
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            s.settimeout(timeout)
            s.connect((self.HOST, self.PORT))
            if len(data) < 2048:
                send(s, BUCKET_SPLIT_TOKEN.join([key.encode('utf-8'), b'set', f'{timeout}'.encode(), data]))
                receive(s)
            else:
                send(s, BUCKET_SPLIT_TOKEN.join([key.encode('utf-8'), b'big-set', f'{timeout}'.encode(), f'{len(data)}'.encode()]))
                receive(s)
                send(s, data)
                receive(s)

            # send(s, key.encode('utf-8'))
            # receive(s)
            #
            # send(s, b'set')
            # receive(s)
            #
            # send(s, data)
            # receive(s)

    @retry_connection
    def get(self, key: str, timeout: float | None = 60 * 5.) -> bytes | None:
        """
        Retrieve data for a given key from the server.

        Args:
            key (str): The key to retrieve data for.
            timeout (float, optional): The timeout for the socket connection in seconds. Defaults to 60 * 5.

        Returns:
            bytes or None: The retrieved data, or None if the key doesn't exist.
        """
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            s.settimeout(timeout)
            s.connect((self.HOST, self.PORT))

            send(s, BUCKET_SPLIT_TOKEN.join([key.encode('utf-8'), b'get', f'{timeout}'.encode(), b'__null__']))

            # send(s, key.encode('utf-8'))
            # receive(s)
            #
            # send(s, b'get')

            data: bytes = receive(s)
            if data[:len(b'__big__')] == b'__big__':
                size = int(data[len(b'__big__'):])
                send(s, b'ok')
                data = receive(s, size)
                return data
            if data == b'__null__':
                return None
            return data

    @retry_connection
    def delete(self, key: str, timeout: float | None = 60 * 5.) -> None:
        """
        Delete data for a given key on the server.

        Args:
            key (str): The key to delete data for.
            timeout (float, optional): The timeout for the socket connection in seconds. Defaults to 60 * 5.
        """
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            s.settimeout(timeout)
            s.connect((self.HOST, self.PORT))

            send(s, BUCKET_SPLIT_TOKEN.join([key.encode('utf-8'), b'delete', f'{timeout}'.encode(), b'__null__']))
            receive(s)

            # send(s, key.encode('utf-8'))
            # receive(s)
            #
            # send(s, b'delete')
            # receive(s)


def check_file_directory(path: str) -> None:
    """
    Ensure the directory for a file path exists, creating it if necessary.

    Args:
        path (str): The file path to check.
    """
    if os.path.dirname(path) == '' or os.path.dirname(path) == '.' or os.path.dirname(path) == '.\\' or os.path.dirname(path) == '/' or os.path.dirname(path) == '\\' or os.path.dirname(path) == './':  # os.path.dirname(path) in {'', '.', '/', '\\', './', '.\\'}:
        return
    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))


def server_get(key: str) -> bytes | None:
    """
    Retrieve data for a given key from the server.

    Args:
        key (str): The key to retrieve data for.

    Returns:
        bytes or None: The retrieved data, or None if the key doesn't exist.
    """
    global database, save_path
    if key not in database:
        if not os.path.exists(os.path.join(save_path, key)):
            return b'__null__'
        with open(os.path.join(save_path, key), 'rb') as f:
            database[key] = f.read()
    return database[key]


def server_set(key: str, data: bytes) -> None:
    """
    Set data for a given key on the server.

    Args:
        key (str): The key to associate with the data.
        data (bytes): The data to store.
    """
    global database, save_path
    database[key] = data
    # check_file_directory(os.path.join(save_path, key))
    # with open(os.path.join(save_path, key), 'wb') as f:
    #     f.write(data)


def server_delete(key: str) -> None:
    """
    Delete data for a given key from the server.

    Args:
        key (str): The key to delete data for.
    """
    global database, save_path
    if key in database:
        del database[key]
    try:
        os.remove(os.path.join(save_path, key))
    except FileNotFoundError:
        pass


def handle_client(conn: socket.socket) -> None:
    """
    Handle a client connection, processing set, get, and delete requests.

    Args:
        conn (socket.socket): The client connection socket.
    """
    global database, save_path
    k: bytes
    m: bytes
    data: bytes
    key: str
    method: str

    def ok():
        send(conn, b'ok')

    with conn:

        k, m, t, data = receive(conn).split(BUCKET_SPLIT_TOKEN)
        key = k.decode('utf-8')
        method = m.decode('utf-8')
        timeout = float(t)

        conn.settimeout(timeout)

        if method == 'big-set':
            size = int(data)
            ok()
            data = receive(conn, size)
            server_set(key, data)
            ok()
        if method == 'set':
            # # ok()
            # # database[key] = data
            # check_file_directory(os.path.join(save_path, key))
            # with open(os.path.join(save_path, key), 'wb') as f:
            #     f.write(data)  # receive(conn))
            server_set(key, data)
            ok()
        elif method == 'get':
            # # if key not in database:
            # #     send(conn, b'__null__')
            # # else:
            # #     send(conn, database[key])
            # if not os.path.exists(os.path.join(save_path, key)):
            #     send(conn, b'__null__')
            # else:
            #     with open(os.path.join(save_path, key), 'rb') as f:
            #         send(conn, f.read())
            data = server_get(key)
            if len(data) < 2048:
                send(conn, server_get(key))
            else:
                send(conn, f'__big__{len(data)}'.encode())
                receive(conn)
                send(conn, data)
        elif method == 'delete':
            # # if key in database:
            # #     del database[key]
            # try:
            #     os.remove(os.path.join(save_path, key))
            # except FileNotFoundError:
            #     pass
            server_delete(key)
            ok()

    if method == 'set' or method == 'big-set':
        check_file_directory(os.path.join(save_path, key))
        with open(os.path.join(save_path, key), 'wb') as f:
            f.write(data)
        database_keys_in_order.append(key)

    while sys.getsizeof(database) > MAX_DATABASE_SIZE:
        key = database_keys_in_order.pop(0) if database_keys_in_order else next(iter(database))
        try:
            del database[key]
        except KeyError:
            pass



class Server:
    """A server for handling bucket storage requests."""
    PORT: int
    HOST: str

    def __init__(self):
        """Initialize the server with host and port from environment variables."""
        self.PORT = BUCKET_SERVER_PORT
        self.HOST = BUCKET_SERVER_HOST

    def loop(self):
        """
        Start the server loop, listening for and handling client connections.
        """
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind((self.HOST, self.PORT))
            s.listen(10)
            while True:
                try:
                    conn, addr = s.accept()
                    conn.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                    conn.settimeout(60. * 60.)

                    threading.Thread(target=handle_client, args=(conn, )).start()
                except TimeoutError as e:
                    pass


def main() -> None:
    """
    Run the bucket server.
    """
    server: Server = Server()
    print('running bucket server', server.HOST, '@', server.PORT)
    server.loop()


if __name__ == '__main__':
    main()

